/* Panorama_Tools	-	Generate, Edit and Convert Panoramic Images
   Copyright (C) 1998,1999 - Helmut Dersch  der@fh-furtwangen.de
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

/*------------------------------------------------------------*/
#include <stdlib.h>
#include <stdint.h>
#include <assert.h>

#include "filter.h"
#include "panolib.h"
#include "utils_math.h"
#include "utils_image.h"

#define DEG_TO_RAD(x) ((x) * 2.0 * M_PI / 360.0 )

// Lookup Tables for Trig-functions and interpolator

#define NATAN 2048
#define NSQRT 2048

static void matrix_matrix_mult( double m1[3][3],double m2[3][3],double result[3][3])
{
	int i,k;
	
	for(i=0;i<3;i++)
		for(k=0; k<3; k++)
			result[i][k] = m1[i][0] * m2[0][k] + m1[i][1] * m2[1][k] + m1[i][2] * m2[2][k];
} /* void matrix_matrix_mult */

/* Set matrix elements based on Euler angles a, b, c */
static void set_transformation_matrix (double m[3][3],
		double a, double b, double c)
{
	double mx[3][3], my[3][3], mz[3][3], dummy[3][3];
	

	// Calculate Matrices;

	mx[0][0] = 1.0 ; 				mx[0][1] = 0.0 ; 				mx[0][2] = 0.0;
	mx[1][0] = 0.0 ; 				mx[1][1] = cos(a) ; 			mx[1][2] = sin(a);
	mx[2][0] = 0.0 ;				mx[2][1] =-mx[1][2] ;			mx[2][2] = mx[1][1];
	
	my[0][0] = cos(b); 				my[0][1] = 0.0 ; 				my[0][2] =-sin(b);
	my[1][0] = 0.0 ; 				my[1][1] = 1.0 ; 				my[1][2] = 0.0;
	my[2][0] = -my[0][2];			my[2][1] = 0.0 ;				my[2][2] = my[0][0];
	
	mz[0][0] = cos(c) ; 			mz[0][1] = sin(c) ; 			mz[0][2] = 0.0;
	mz[1][0] =-mz[0][1] ; 			mz[1][1] = mz[0][0] ; 			mz[1][2] = 0.0;
	mz[2][0] = 0.0 ;				mz[2][1] = 0.0 ;				mz[2][2] = 1.0;

	/* Calculate `m = mz * mx * my' */

	matrix_matrix_mult( mz,	mx,	dummy);
	matrix_matrix_mult( dummy, my, m);
} /* void SetMatrix */

#if 0
static double lanczos (double x)
{
	static double table[257];

	assert ((x >= 0.0) && (x <= 1.0));

	if (table[0] != 1.0)
	{
		int i;

		table[0] = 1.0;
		for (i = 1; i <= 256; i++)
		{
			double t = ((double) i) / 256.0;
			table[i] = 2 * sin (M_PI * t) * sin (M_PI_2 * t) / (M_PI * M_PI * t * t);
		}
	}

	return (table[(int) (0.5 + 256.0 * x)]);
} /* double lanczos */
#endif

static int copy_pixel (ui_image_t *dest, const ui_image_t *src,
		int x_dest, int y_dest,
		double x_src_fp, double y_src_fp,
		interpolator_t interp)
{
	uint32_t pixel_dest = (y_dest * dest->width) + x_dest;

	assert (x_src_fp >= 0.0);

	if (y_src_fp < 0.0)
		y_src_fp = 0.0;

	if (interp == BILINEAR)
	{
		int x_src_left;
		int x_src_right;
		int y_src_top;
		int y_src_bottom;

		double x_right_frac;
		double y_bottom_frac;

		int i;

		x_src_left = (int) x_src_fp;
		x_src_right = (x_src_left + 1) % src->width;

		y_src_top  = (int) y_src_fp;
		y_src_bottom = y_src_top + 1;
		if (y_src_bottom >= src->height)
			y_src_bottom = src->height - 1;

		x_right_frac = x_src_fp - x_src_left;
		y_bottom_frac = y_src_fp - y_src_top;

		/*
		 * The polynomial function
		 *   2x^3 - 3x^2 + 1
		 * is similar to the Lanczos function, but much easier to compute. It's
		 * applied here to improve sharpness. 
		 */
		{
			double t = x_src_fp - x_src_left;
			x_right_frac = -2.0 * t * t * t + 3.0 * t * t;

			t = y_src_fp - y_src_top;
			y_bottom_frac = -2.0 * t * t * t + 3.0 * t * t;
		}

		assert ((x_right_frac >= 0.0) && (x_right_frac <= 1.0));
		assert ((y_bottom_frac >= 0.0) && (y_bottom_frac <= 1.0));

		for (i = 0; i < 3; i++)
		{
			uint8_t values[2][2];
			double value_left;
			double value_right;
			double value_final;

			values[0][0] = src->data[i][(y_src_top * src->width) + x_src_left];
			values[0][1] = src->data[i][(y_src_top * src->width) + x_src_right];
			values[1][0] = src->data[i][(y_src_bottom * src->width) + x_src_left];
			values[1][1] = src->data[i][(y_src_bottom * src->width) + x_src_right];

			value_left = (1.0 - y_bottom_frac) * values[0][0]
				+ y_bottom_frac * values[1][0];
			value_right = (1.0 - y_bottom_frac) * values[0][1]
				+ y_bottom_frac * values[1][1];

			value_final = (1.0 - x_right_frac) * value_left + x_right_frac * value_right;

			assert ((value_final >= 0.0) && (value_final <= 255.0));

			dest->data[i][pixel_dest] = (uint8_t) (value_final + 0.5);
		}
	}
	else /* if (interp == NNEIGHBOUR) */
	{
		int x_src = (int) (x_src_fp + 0.5) % src->width;
		int y_src = (int) (y_src_fp + 0.5) % src->height;

		if ((x_src < 0) || (x_src >= src->width)
				|| (y_src < 0) || (y_src >= src->height))
		{
			dest->data[0][pixel_dest] = 0;
			dest->data[1][pixel_dest] = 0;
			dest->data[2][pixel_dest] = 0;
		}
		else
		{
			uint32_t pixel_src = (y_src * src->width) + x_src;

			dest->data[0][pixel_dest] = src->data[0][pixel_src];
			dest->data[1][pixel_dest] = src->data[1][pixel_src];
			dest->data[2][pixel_dest] = src->data[2][pixel_src];
		}
	}

	return (0);
} /* int copy_pixel */

//    Main transformation function. Destination image is calculated using transformation
//    Function "func". Either all colors (color = 0) or one of rgb (color =1,2,3) are
//    determined. If successful, TrPtr->success = 1. Memory for destination image
//    must have been allocated and locked!

int pl_extract_view (ui_image_t *view, const ui_image_t *pano,
		double pitch, double yaw, double fov, interpolator_t interp)
{
	int x_dest, y_dest; // Loop through destination image

	int dest_width_left = view->width  / 2 ;  
	int dest_height_top = view->height / 2 ;
	int src_width_left =  pano->width   / 2 ;
	int src_height_top =  pano->height  / 2 ;
	
	double v[3];
	int	x_min, x_max, y_min, y_max;

	/* The transformation matrix */
	double tm[3][3];

	double dist_r;
	double dist_e;

	{ /* What the fuck does this? -octo */
		double a = DEG_TO_RAD (fov);
		double b = 2.0 * M_PI; /* DEG_TO_RAD (360.0) */

		dist_r = ((double) view->width) / (2.0 * tan (a / 2.0));
		dist_e = ((double) pano->width) / b;
	}

	set_transformation_matrix (tm, DEG_TO_RAD (pitch), DEG_TO_RAD (yaw), 0.0);

	x_min = -dest_width_left; x_max = view->width - dest_width_left;
	y_min = -dest_height_top; y_max = view->height - dest_height_top;

	for(y_dest = y_min; y_dest < y_max; y_dest++)
	{
		for(x_dest = x_min; x_dest < x_max; x_dest++)
		{
			double x_src_fp;
			double y_src_fp;

			v[0] = tm[0][0] * x_dest + tm[1][0] * y_dest + tm[2][0] * dist_r;
			v[1] = tm[0][1] * x_dest + tm[1][1] * y_dest + tm[2][1] * dist_r;
			v[2] = tm[0][2] * x_dest + tm[1][2] * y_dest + tm[2][2] * dist_r;

			x_src_fp = dist_e * atan2 (v[0], v[2]);
			y_src_fp = dist_e * atan2 (v[1], sqrt (v[2] * v[2] + v[0] * v[0]));

			copy_pixel (view, pano,
					dest_width_left + x_dest, dest_height_top + y_dest,
					src_width_left + x_src_fp, src_height_top + y_src_fp,
					interp); /* FIXME */
		}
	}
	
	return (0);
}

/*
 * vim: set tabstop=4 softtabstop=4 shiftwidth=4 :
 */
